import datetime
from collections import OrderedDict

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from .models import Schedule


class MultiDatetimeField(forms.Field):
    widget = forms.MultipleHiddenInput

    def _clean_item(self, item):
        if isinstance(item, tuple):
            return item

        try:
            dt = datetime.datetime.strptime(item, '%Y-%m-%d %H:%M')
        except (ValueError, TypeError):
            raise ValidationError(_('Invalid datetime format'), code='invalid_datetime')
        return dt.date(), dt.time()

    def clean(self, value):
        if not value:
            return []
        elif not isinstance(value, (list, tuple)):
            raise ValidationError(_('Enter a list of values'), code='invalid_list')

        return [self._clean_item(item) for item in set(value)]

    def prepare_value(self, value):
        data = self.clean(super().prepare_value(value))

        if data:
            dates, times = zip(*data)
            dates = sorted(set(dates))
            times = sorted(set(times))
        else:
            dates = []
            times = []

        table = OrderedDict()
        for date in dates:
            table[date] = OrderedDict()
            for time in times:
                table[date][time] = False
        for (date, time) in data:
            table[date][time] = True

        return {
            'timeslots': table,
            'times': times,
        }


class ScheduleForm(forms.ModelForm):
    timeslots = MultiDatetimeField()

    class Meta:
        model = Schedule
        fields = ['title', 'advance', 'owners']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance and self.instance.pk:
            self.fields['timeslots'].initial = list(
                self.instance.timeslot_set.values_list('date', 'time')
            )

    def save_timeslots(self, schedule):
        tuples = self.cleaned_data['timeslots']
        old = list(schedule.timeslot_set.all())
        old_tuples = [(t.date, t.time) for t in old]

        for timeslot in old:
            if (timeslot.date, timeslot.time) not in tuples:
                timeslot.delete()

        for date, time in tuples:
            if (date, time) not in old_tuples:
                schedule.timeslot_set.create(date=date, time=time)

    def save(self):
        schedule = super().save()
        self.save_timeslots(schedule)
        return schedule
